/*
This file is part of Yama.

Yama is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Yama is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Yama.  If not, see <https://www.gnu.org/licenses/>.
*/

use std::sync::Arc;
use std::thread;
use std::thread::JoinHandle;

use anyhow::anyhow;
use crossbeam_channel::{Receiver, Sender};
use log::error;
use zstd::block::{Compressor, Decompressor};

use crate::pile::{Keyspace, RawPile};

pub const DECOMPRESS_CAPACITY: usize = 32 * 1024 * 1024;

#[derive(Clone)]
pub struct CompressionSettings {
    /// Raw dictionary to pass to Zstd for compression and decompression
    pub dictionary: Arc<Vec<u8>>,
    /// The compression level, passed to Zstd.
    pub level: i32,
    /// The number of compressor threads to use.
    pub num_compressors: u32,
    /// The number of decompressor threads to use.
    pub num_decompressors: u32,
}

pub struct RawPileCompressor<R: RawPile> {
    underlying: R,
    compressor: Sender<(Vec<u8>, Sender<Vec<u8>>)>,
    decompressor: Sender<(Vec<u8>, Sender<Vec<u8>>)>,
}

impl<R: RawPile> RawPileCompressor<R> {
    pub fn new(
        underlying: R,
        settings: CompressionSettings,
    ) -> anyhow::Result<(Self, Vec<JoinHandle<()>>)> {
        let (com_s, com_r) = crossbeam_channel::bounded(4);
        let (dec_s, dec_r) = crossbeam_channel::bounded(4);

        let mut handles = Vec::new();

        for worker in 0..settings.num_compressors {
            let settings = settings.clone();
            let com_r = com_r.clone();

            let builder = thread::Builder::new().name(format!("yama compressor {}", worker));

            handles.push(builder.spawn(move || {
                if let Err(e) = Self::compressor_worker(com_r, settings) {
                    error!("compressor worker failed: {:?}", e);
                }
            })?);
        }

        for worker in 0..settings.num_decompressors {
            let settings = settings.clone();
            let dec_r = dec_r.clone();

            let builder = thread::Builder::new().name(format!("yama decompressor {}", worker));

            handles.push(builder.spawn(move || {
                if let Err(e) = Self::decompressor_worker(dec_r, settings) {
                    error!("decompressor worker failed: {:?}", e);
                }
            })?);
        }

        Ok((
            RawPileCompressor {
                underlying,
                compressor: com_s,
                decompressor: dec_s,
            },
            handles,
        ))
    }

    fn compressor_worker(
        queue: Receiver<(Vec<u8>, Sender<Vec<u8>>)>,
        settings: CompressionSettings,
    ) -> anyhow::Result<()> {
        let mut compressor = Compressor::with_dict(settings.dictionary.as_ref().clone());
        while let Ok((job, response_sender)) = queue.recv() {
            let result = compressor.compress(&job, settings.level)?;
            response_sender
                .send(result)
                .or(Err(anyhow!("Couldn't send compression result")))?;
        }
        Ok(())
    }

    fn decompressor_worker(
        queue: Receiver<(Vec<u8>, Sender<Vec<u8>>)>,
        settings: CompressionSettings,
    ) -> anyhow::Result<()> {
        let mut decompressor = Decompressor::with_dict(settings.dictionary.as_ref().clone());
        while let Ok((job, response_sender)) = queue.recv() {
            let result = decompressor.decompress(&job, DECOMPRESS_CAPACITY)?;
            response_sender
                .send(result)
                .or(Err(anyhow!("Couldn't send decompression result")))?;
        }
        Ok(())
    }

    fn decompress(&self, data: &[u8]) -> anyhow::Result<Vec<u8>> {
        let (ret_s, ret_r) = crossbeam_channel::bounded(0);
        self.decompressor
            .send((data.to_vec(), ret_s))
            .or(Err(anyhow!("couldn't send to decompressor")))?;

        Ok(ret_r.recv().or(Err(anyhow!("couldn't receive result")))?)
    }

    fn compress(&self, compressed_data: &[u8]) -> anyhow::Result<Vec<u8>> {
        let (ret_s, ret_r) = crossbeam_channel::bounded(0);
        self.compressor
            .send((compressed_data.to_vec(), ret_s))
            .or(Err(anyhow!("couldn't send to compressor")))?;

        Ok(ret_r.recv().or(Err(anyhow!("couldn't receive result")))?)
    }
}

impl<R: RawPile> RawPile for RawPileCompressor<R> {
    fn exists(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<bool> {
        self.underlying.exists(kind, key)
    }

    fn read(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<Option<Vec<u8>>> {
        if let Some(data) = self.underlying.read(kind, key)? {
            Ok(Some(self.decompress(&data)?))
        } else {
            Ok(None)
        }
    }
    fn write(&self, kind: Keyspace, key: &[u8], value: &[u8]) -> anyhow::Result<()> {
        let compressed = self.compress(value)?;
        self.underlying.write(kind, key, &compressed)
    }

    fn delete(&self, kind: Keyspace, key: &[u8]) -> anyhow::Result<()> {
        self.underlying.delete(kind, key)
    }

    fn list_keys(
        &self,
        kind: Keyspace,
    ) -> anyhow::Result<Box<dyn Iterator<Item = anyhow::Result<Vec<u8>>>>> {
        self.underlying.list_keys(kind)
    }
    fn flush(&self) -> anyhow::Result<()> {
        self.underlying.flush()
    }
    fn check_lowlevel(&self) -> anyhow::Result<bool> {
        self.underlying.check_lowlevel()
    }
}
